#!/usr/bin/env python3
"""Index datasets found from an SQS queue into Postgres"""

import concurrent
import json
import logging
import sys
from typing import Any, Dict, Optional, Tuple

import click
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.ui.click import environment_option, pass_config
from odc.apps.dc_tools.utils import (
    DatasetExists,
    allow_unsafe,
    archive_less_mature,
    bbox,
    index_update_dataset,
    item_to_meta_uri,
    limit,
    publish_action,
    rename_product,
    url_string_replace,
    statsd_gauge_reporting,
    statsd_setting,
    update_if_exists_flag,
    skip_lineage,
    fail_on_missing_lineage,
    verify_lineage,
)
from pystac.item import Item
from pystac_client import Client
from sqlalchemy.exc import OperationalError, ProgrammingError

logging.basicConfig(
    level=logging.WARNING,
    format="%(asctime)s: %(levelname)s: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S",
)


def _parse_options(options: Optional[str]) -> Dict[str, Any]:
    parsed_options = {}

    if options is not None:
        for option in options.split("#"):
            try:
                key, value = option.split("=")

                try:
                    value = json.loads(value)
                except Exception:  # pylint:disable=broad-except
                    logging.warning(
                        "Failed to handle value %s for key %s as JSON, using str",
                        value,
                        key,
                    )
                parsed_options[key] = value
            except Exception:  # pylint:disable=broad-except
                logging.warning(
                    "Couldn't parse option %s format is key=value",
                    option,
                    exc_info=True,
                )

    return parsed_options


def process_item(
    item: Item,
    dc: Datacube,
    update_if_exists: bool,
    allow_unsafe: bool,
    rename_product: Optional[str] = None,
    url_string_replace: tuple[str, str] | None = None,
    archive_less_mature: int | None = None,
    publish_action: bool = False,
    **kwargs,
) -> None:
    dataset, uri, stac = item_to_meta_uri(item, dc, rename_product, url_string_replace)
    if uri is None:
        raise ValueError(
            f"The links field did not contain a self-reference for item {item}"
        )
    doc2ds = Doc2Dataset(dc.index, **kwargs)
    index_update_dataset(
        dataset,
        uri,
        dc,
        doc2ds,
        update_if_exists=update_if_exists,
        allow_unsafe=allow_unsafe,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        stac_doc=stac,
    )


def stac_api_to_odc(
    dc: Datacube,
    update_if_exists: bool,
    config: dict,
    catalog_href: str,
    allow_unsafe: bool = True,
    rename_product: Optional[str] = None,
    url_string_replace: tuple[str, str] | None = None,
    archive_less_mature: int | None = None,
    publish_action: Optional[str] = None,
    **kwargs,
) -> Tuple[int, int, int]:
    client = Client.open(catalog_href)

    search = client.search(**config)
    n_items = search.matched()
    if n_items is not None:
        logging.info("Found %s items to index", n_items)
        if n_items == 0:
            logging.warning("Didn't find any items, finishing.")
            return 0, 0, 0
    else:
        logging.warning("API did not return the number of items.")

    # Do the indexing of all the things
    success = 0
    failure = 0
    skipped = 0

    sys.stdout.write("\rIndexing from STAC API...\n")
    with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
        future_to_item = {
            executor.submit(
                process_item,
                item,
                dc,
                update_if_exists=update_if_exists,
                allow_unsafe=allow_unsafe,
                rename_product=rename_product,
                url_string_replace=url_string_replace,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
                **kwargs,
            ): item.id
            for item in search.items()
        }
        for future in concurrent.futures.as_completed(future_to_item):
            item = future_to_item[future]
            try:
                _ = future.result()
                success += 1
                if success % 10 == 0:
                    sys.stdout.write(f"\rAdded {success} datasets...")
            except DatasetExists:
                skipped += 1
            except Exception:  # pylint:disable=broad-except
                logging.exception("Failed to handle item %s", item)
                failure += 1
    sys.stdout.write("\r")

    return success, failure, skipped


@click.command("stac-to-dc")
@environment_option
@pass_config
@limit
@update_if_exists_flag
@allow_unsafe
@click.option(
    "--catalog-href",
    type=str,
    default="https://earth-search.aws.element84.com/v0/",
    help="URL to the catalog to search",
)
@click.option(
    "--collections",
    type=str,
    default=None,
    help="Comma separated list of collections to search",
)
@bbox
@click.option(
    "--datetime",
    type=str,
    default=None,
    help="Dates to search, either one day or an inclusive range, e.g. 2020-01-01 or 2020-01-01/2020-01-02",
)
@click.option(
    "--options",
    type=str,
    default=None,
    help="Other search terms, as a # separated list, i.e., --options=cloud_cover=0,100#sky=green",
)
@rename_product
@url_string_replace
@archive_less_mature
@publish_action
@statsd_setting
@skip_lineage
@fail_on_missing_lineage
@verify_lineage
def cli(
    cfg_env,
    limit,
    update_if_exists,
    allow_unsafe,
    catalog_href,
    collections,
    bbox,
    datetime,
    options,
    rename_product,
    url_string_replace,
    archive_less_mature,
    publish_action,
    statsd_setting,
    skip_lineage,
    fail_on_missing_lineage,
    verify_lineage,
) -> None:
    """
    Iterate through STAC items from a STAC API and add them to datacube.
    """
    config = _parse_options(options)

    # Format the search terms
    if bbox:
        config["bbox"] = list(map(float, bbox.split(",")))

    if collections:
        config["collections"] = collections.split(",")

    if datetime:
        config["datetime"] = datetime

    if url_string_replace:
        url_string_replace_tuple = tuple(url_string_replace.split(","))
        if len(url_string_replace_tuple) != 2:
            raise ValueError(
                "url_string_replace must be two strings separated by a comma"
            )
    else:
        url_string_replace_tuple = None

    # Always set the limit, because some APIs will stop at an arbitrary
    # number if max_items is not None.
    config["max_items"] = limit

    # Do the thing
    try:
        dc = Datacube(env=cfg_env, app="stac-api-to-dc")
    except (OperationalError, ProgrammingError) as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)
    added, failed, skipped = stac_api_to_odc(
        dc,
        update_if_exists,
        config,
        catalog_href,
        allow_unsafe=allow_unsafe,
        rename_product=rename_product,
        url_string_replace=url_string_replace_tuple,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        skip_lineage=skip_lineage,
        fail_on_missing_lineage=fail_on_missing_lineage,
        verify_lineage=verify_lineage,
    )

    print(
        f"Added {added} Datasets, failed {failed} Datasets, skipped {skipped} Datasets"
    )
    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:stac_api_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:stac_api_to_dc", "action:failed"], statsd_setting
        )
        statsd_gauge_reporting(
            skipped, ["app:stac_api_to_dc", "action:skipped"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



import click
import sys
from odc.aio import S3Fetcher, s3_find_glob


@click.command("s3-find")
@click.option(
    "--skip-check",
    is_flag=True,
    help="Assume file exists when listing exact file rather than wildcard.",
)
@click.option("--no-sign-request", is_flag=True, help="Do not sign AWS S3 requests")
@click.option(
    "--request-payer",
    is_flag=True,
    help="Needed when accessing requester pays public buckets",
)
@click.argument("uri", type=str, nargs=1)
def cli(uri, skip_check, no_sign_request=None, request_payer=False) -> None:
    """List files on S3 bucket.

    Example:

       \b
       List files in directory that match `*yaml`
        > s3-find 's3://mybucket/some/path/*yaml'

       \b
       List files in directory and all sub-directories that match `*yaml`
        > s3-find 's3://mybucket/some/path/**/*yaml'

       \b
       List files that match `*yaml` 2 levels deep from known path
        > s3-find 's3://mybucket/some/path/*/*/*yaml'

       \b
       List directories 2 levels deep from known path
        > s3-find 's3://mybucket/some/path/*/*/'

       \b
       List all files named `metadata.yaml` 2 directories deep
        > s3-find 's3://mybucket/some/path/*/*/metadata.yaml'
    """
    flush_freq = 100

    opts = {}
    if request_payer:
        opts["RequestPayer"] = "requester"

    s3 = S3Fetcher(aws_unsigned=no_sign_request)

    try:
        stream = s3_find_glob(uri, skip_check=skip_check, s3=s3, **opts)
        for i, o in enumerate(stream):
            print(
                o.url, flush=(i % flush_freq == 0)
            )  # pylint:disable=superfluous-parens
    except ValueError as ve:
        click.echo(str(ve), err=True)
        sys.exit(1)
    except Exception as e:  # pylint:disable=broad-except
        click.echo(str(e), err=True)
        sys.exit(1)


if __name__ == "__main__":
    cli()



"""
Test for SQS to DC tool
"""

import boto3
import json
import os
import pytest
from deepdiff import DeepDiff
from functools import partial
from moto import mock_aws
from odc.aws.queue import get_messages
from pathlib import Path
from pprint import pformat

from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.utils import documents
from odc.apps.dc_tools._stac import stac_transform
from odc.apps.dc_tools.sqs_to_dc import (
    extract_metadata_from_message,
    handle_bucket_notification_message,
    handle_json_message,
)
from odc.apps.dc_tools.utils import index_update_dataset

record_message = {
    "Records": [
        {
            "eventVersion": "2.1",
            "eventSource": "aws:s3",
            "awsRegion": "us-east-2",
            "eventTime": "2018-12-19T01:51:03.251Z",
            "eventName": "ObjectCreated:Put",
            "userIdentity": {"principalId": "AWS:AIDAIZLCFC5TZD36YHNZY"},
            "requestParameters": {"sourceIPAddress": "52.46.82.38"},
            "responseElements": {
                "x-amz-request-id": "6C05F1340AA50D21",
                "x-amz-id-2": "9e8KovdAUJwmYu1qnEv+urrO8T0vQ+UOpkPnFYLE6agmJSn745/T3/tVs0Low/vXonTdATvW23M=",
            },
            "s3": {
                "s3SchemaVersion": "1.0",
                "configurationId": "test_SQS_Notification_1",
                "bucket": {
                    "name": "dea-public-data",
                    "ownerIdentity": {"principalId": "A2SGQBYRFBZET"},
                    "arn": "arn:aws:s3:::dea-public-data",
                },
                "object": {
                    "key": "cemp_insar/insar/displacement/alos/2009/06/17/alos_cumul_2009-06-17.yaml",
                    "size": 713,
                    "eTag": "1ff1209e4140b4ff7a9d2b922f57f486",
                    "sequencer": "005C19A40717D99642",
                },
            },
        }
    ]
}

sqs_message = {
    "Type": "Notification",
    "MessageId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxxxxxxxxx",
    "TopicArn": "arn:aws:sns:ap-southeast-2:xxxxxxxxxxxxxxxxx:DEANewData",
    "Subject": "Amazon S3 Notification",
    "Message": json.dumps(record_message),
    "Timestamp": "2020-08-21T08:28:45.921Z",
    "SignatureVersion": "1",
    "Signature": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    "SigningCertURL": "https://sns.ap-southeast-2.amazonaws.com/SimpleNotificationService-"
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx.pem",
    "UnsubscribeURL": "https://sns.ap-southeast-2.amazonaws.com/?Action=Unsubscribe"
    "&SubscriptionArn=arn:aws:sns:ap-southeast-2:xxxxxxxxxxxxxxx:DEANewData:"
    "xxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxxxx",
}


TEST_DATA_FOLDER: Path = Path(__file__).parent / "data"
LANDSAT_C3_SQS_MESSAGE: str = (
    "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.stac-item.sqs.json"
)
LANDSAT_C3_ODC_YAML: str = (
    "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.odc-metadata.sqs.yaml"
)
SENTINEL_2_NRT_MESSAGE: str = "sentinel-2-nrt_2020_08_21.json"

deep_diff = partial(
    DeepDiff, significant_digits=6, ignore_type_in_groups=[(tuple, list)]
)


@pytest.fixture
def aws_credentials() -> None:
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"


def test_extract_metadata_from_message(
    aws_credentials, odc_test_db_with_products, cfg_env
) -> None:
    with mock_aws():
        TEST_QUEUE_NAME = "a_test_queue"
        sqs_resource = boto3.resource("sqs")

        dc = Datacube(env=cfg_env)

        a_queue = sqs_resource.create_queue(QueueName=TEST_QUEUE_NAME)
        assert int(a_queue.attributes.get("ApproximateNumberOfMessages")) == 0

        a_queue.send_message(MessageBody=json.dumps(sqs_message))
        assert int(a_queue.attributes.get("ApproximateNumberOfMessages")) == 1

        assert dc.index.datasets.get("69a6eca2-ca45-4808-a5b3-694029200c43") is None

        queue = sqs_resource.get_queue_by_name(QueueName=TEST_QUEUE_NAME)

        msg = next(get_messages(queue))
        metadata = extract_metadata_from_message(msg)
    data, uri = handle_bucket_notification_message(
        msg, metadata, "cemp_insar/insar/displacement/alos/*", True
    )

    assert (
        uri
        == "s3://dea-public-data/cemp_insar/insar/displacement/alos/2009/06/17/alos_cumul_2009-06-17.yaml"
    )
    assert isinstance(data, dict)

    doc2ds = Doc2Dataset(dc.index, products=["cemp_insar_alos_displacement"])
    index_update_dataset(
        data,
        uri,
        dc,
        doc2ds,
    )

    assert dc.index.datasets.get("69a6eca2-ca45-4808-a5b3-694029200c43") is not None


def test_handle_json_message(ga_ls8c_ard_3_message, ga_ls8c_ard_3_yaml) -> None:
    actual_doc, uri = handle_json_message(
        ga_ls8c_ard_3_message, "STAC-LINKS-REL:odc_yaml"
    )

    assert type(actual_doc) is dict
    assert ga_ls8c_ard_3_yaml["id"] == actual_doc["id"]
    # actual_doc comes from DEA S3 bucket and may have CRS in lower case.
    assert ga_ls8c_ard_3_yaml["crs"].upper() == actual_doc["crs"].upper()
    assert ga_ls8c_ard_3_yaml["product"]["name"] == actual_doc["product"]["name"]
    assert ga_ls8c_ard_3_yaml["label"] == actual_doc["label"]

    # Test geometry field
    doc_diff = deep_diff(ga_ls8c_ard_3_yaml["geometry"], actual_doc["geometry"])
    assert doc_diff == {}, pformat(doc_diff)

    # Test grids field
    doc_diff = deep_diff(
        ga_ls8c_ard_3_yaml["grids"]["default"], actual_doc["grids"]["default"]
    )
    assert doc_diff == {}, pformat(doc_diff)

    # Test measurements field
    doc_diff = deep_diff(
        ga_ls8c_ard_3_yaml["measurements"],
        actual_doc["measurements"],
        exclude_paths=["root['nbart_panchromatic']"],
    )
    assert doc_diff == {}, pformat(doc_diff)


def test_odc_metadata_link(ga_ls8c_ard_3_message) -> None:
    actual_doc, uri = handle_json_message(
        ga_ls8c_ard_3_message, "STAC-LINKS-REL:odc_yaml"
    )
    assert (
        uri == "https://dea-public-data.s3-ap-southeast-2.amazonaws.com/"
        "baseline/ga_ls8c_ard_3/088/080/2020/05/25/"
        "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.odc-metadata.yaml"
    )


def test_stac_link(ga_ls8c_ard_3_message) -> None:
    metadata, uri = handle_json_message(ga_ls8c_ard_3_message, None)
    assert (
        uri != "https://dea-public-data.s3-ap-southeast-2.amazonaws.com/"
        "baseline/ga_ls8c_ard_3/088/080/2020/05/25/"
        "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.odc-metadata.yaml"
    )
    assert (
        uri == "https://dea-public-data.s3-ap-southeast-2.amazonaws.com/"
        "baseline/ga_ls8c_ard_3/088/080/2020/05/25/"
        "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.stac-item.json"
    )


def test_transform(ga_ls8c_ard_3_message, ga_ls8c_ard_3_yaml) -> None:
    actual_doc, uri = handle_json_message(ga_ls8c_ard_3_message, None)
    actual_doc = stac_transform(actual_doc)

    assert ga_ls8c_ard_3_yaml["id"] == actual_doc["id"]
    assert ga_ls8c_ard_3_yaml["crs"] == actual_doc["crs"]
    assert ga_ls8c_ard_3_yaml["product"]["name"] == actual_doc["product"]["name"]
    assert ga_ls8c_ard_3_yaml["label"] == actual_doc["label"]

    # Test geometry field
    # TODO: fix geometry test here.
    # doc_diff = deep_diff(ga_ls8c_ard_3_yaml["geometry"], actual_doc["geometry"])
    # assert doc_diff == {}, pformat(doc_diff)

    # Test grids field
    doc_diff = deep_diff(ga_ls8c_ard_3_yaml["grids"], actual_doc["grids"])
    assert doc_diff == {}, pformat(doc_diff)

    # Test measurements field
    doc_diff = deep_diff(ga_ls8c_ard_3_yaml["measurements"], actual_doc["measurements"])
    assert doc_diff == {}, pformat(doc_diff)

    # Test properties field
    doc_diff = deep_diff(
        ga_ls8c_ard_3_yaml["properties"],
        actual_doc["properties"],
        exclude_paths=[
            "root['odc:product']",
            "root['datetime']",
            "root['dtr:start_datetime']",
            "root['dtr:end_datetime']",
            "root['odc:processing_datetime']",
            "root['proj:epsg']",
            "root['proj:shape']",
            "root['proj:transform']",
        ],
    )
    assert doc_diff == {}, pformat(doc_diff)


@pytest.fixture
def ga_ls8c_ard_3_message():
    with TEST_DATA_FOLDER.joinpath(LANDSAT_C3_SQS_MESSAGE).open("r") as f:
        body = json.load(f)
    metadata = json.loads(body["Message"])
    return metadata


@pytest.fixture
def ga_ls8c_ard_3_yaml():
    metadata = yield from documents.load_documents(
        TEST_DATA_FOLDER.joinpath(LANDSAT_C3_ODC_YAML)
    )
    return metadata



import click
import logging
import sys
from odc.aws.queue import redrive_queue


@click.command("redrive-to-queue")
@click.argument("queue", required=True)
@click.argument("to-queue", required=False)
@click.option(
    "--limit",
    "-l",
    help="Limit the number of messages to transfer.",
    default=None,
)
@click.option(
    "--dryrun", is_flag=True, default=False, help="Don't actually do real work"
)
def cli(queue, to_queue, limit, dryrun):
    """
    Redrives all the messages from the given sqs queue to the destination
    """

    logging.basicConfig(
        level=logging.INFO,
        format="[%(asctime)s] %(levelname)s - %(message)s",
        stream=sys.stdout,
    )

    _log = logging.getLogger(__name__)

    if limit is not None:
        try:
            limit = int(limit)
        except ValueError:
            raise ValueError(f"Limit {limit} is not valid")

        if limit < 1:
            raise ValueError(f"Limit {limit} is not valid.")

    count = redrive_queue(queue, to_queue, limit, dryrun, max_wait=1)

    if count == 0:
        _log.info("No messages to redrive")
        return

    if not dryrun:
        _log.info("Completed sending %s messages to the queue", count)
    else:
        _log.warning(
            "DRYRUN enabled, would have pushed approx %s messages to the queue", count
        )


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Build S3 iterators using odc-tools
and index datasets found into RDS
"""

import logging
import sys
from types import SimpleNamespace
from typing import Any, Dict

import botocore
import click
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.ui.click import environment_option, pass_config
from datacube.utils.aws import (
    _aws_unsigned_check_env,
    auto_find_region,
    s3_client,
    s3_fetch,
)
from datacube.utils.documents import parse_doc_stream
from odc.aio import S3Fetcher, s3_find_glob
from odc.apps.dc_tools.utils import (
    DatasetExists,
    IndexingError,
    allow_unsafe,
    archive_less_mature,
    fail_on_missing_lineage,
    index_update_dataset,
    item_to_meta_uri,
    get_self_link,
    no_sign_request,
    publish_action,
    rename_product,
    request_payer,
    skip_check,
    skip_lineage,
    statsd_gauge_reporting,
    statsd_setting,
    transform_stac,
    update_flag,
    update_if_exists_flag,
    url_string_replace,
    verify_lineage,
)
from pystac import Item
from sqlalchemy.exc import OperationalError, ProgrammingError


def doc_error(uri, doc, exc: Exception) -> None:
    """Log the internal errors parsing docs"""
    logging.exception("Failed to parse doc at %s", uri)


class SimpleFetcher:
    """
    Super simple S3 URL fetcher.

    Args:
        region_name (str, optional): AWS region name to use for S3 requests.If not provided, attempts to auto-detect.
        aws_unsigned (bool, optional): If True, disables AWS request signing for public buckets.
        request_opts (dict, optional): Additional options to pass to the S3 fetch operation.

    Methods:
        __call__(uris):
            Fetches a sequence of S3 URLs.
            Args:
                uris (Iterable): Sequence of S3 URLs.
            Yields:
                SimpleNamespace: For each input, yields an object with:
                    - url (str): The S3 URL.
                    - data (bytes or None): The fetched data.
            Notes:
                - The order of results is not guaranteed to match the input order.
                - One result is yielded for each input URI.
    """

    def __init__(
        self,
        region_name: str | None = None,
        aws_unsigned: bool | None = None,
        request_opts: Dict[Any, Any] | None = None,
    ):
        opts = {}

        if request_opts is None:
            request_opts = {}

        if region_name is None:
            region_name = auto_find_region()

        if aws_unsigned is None:
            aws_unsigned = _aws_unsigned_check_env()

        if aws_unsigned:
            opts["signature_version"] = botocore.UNSIGNED

        opts["region_name"] = region_name
        opts["aws_unsigned"] = aws_unsigned

        self.opts = opts
        self.request_opts = request_opts

    def __call__(self, uris):
        for url in uris:
            client = s3_client(**self.opts)
            data = s3_fetch(s3=client, url=url, **self.request_opts)

            yield SimpleNamespace(url=url, data=data)


@click.command("s3-to-dc")
@environment_option
@pass_config
@click.option(
    "--log",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    default="WARNING",
    show_default=True,
    help="control the log level, e.g., --log=error",
)
@skip_lineage
@fail_on_missing_lineage
@verify_lineage
@transform_stac
@update_flag
@update_if_exists_flag
@allow_unsafe
@skip_check
@no_sign_request
@statsd_setting
@request_payer
@archive_less_mature
@publish_action
@rename_product
@url_string_replace
@click.option(
    "--convert-bools",
    is_flag=True,
    default=False,
    help="Convert boolean properties to strings for backwards compatibility",
)
@click.argument("uris", nargs=-1)
@click.argument("product", type=str, nargs=1, required=False)
def cli(
    cfg_env,
    log: str,
    skip_lineage: bool,
    fail_on_missing_lineage: bool,
    verify_lineage: bool,
    stac: bool,
    update: bool,
    update_if_exists: bool,
    allow_unsafe: bool,
    skip_check: bool,
    no_sign_request: bool,
    statsd_setting: str,
    request_payer: bool,
    archive_less_mature: int | None,
    publish_action: str,
    rename_product: str | None,
    url_string_replace: str | None,
    convert_bools: bool,
    uris: list[str],
    product: str,
) -> None:
    """
    Iterate through files in an S3 bucket and add them to datacube.

    File uris can be provided as a glob, or as a list of absolute URLs.
    If more than one uri is given, all will be treated as absolute URLs.

    Product is optional; if one is provided, it must match all datasets.
    Can provide a single product name or a space separated list of multiple products
    (formatted as a single string).
    """
    log_level = getattr(logging, log.upper())
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s: %(levelname)s: %(message)s",
        datefmt="%m/%d/%Y %I:%M:%S",
    )

    opts = {}
    if request_payer:
        opts["RequestPayer"] = "requester"

    try:
        dc = Datacube(env=cfg_env, app="s3-to-dc")
    except (OperationalError, ProgrammingError) as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)
    # if it's a uri, a product wasn't provided, and 'product' is actually another uri
    if product.startswith("s3://"):
        candidate_products = []
        uris += (product,)
    else:
        # Check datacube connection and products
        candidate_products = product.split()
        odc_products = set(dc.list_products().name.values)
        if not set(candidate_products).issubset(odc_products):
            missing_products = list(set(candidate_products) - odc_products)
            print(
                f"Error: Requested Product/s {', '.join(missing_products)} "
                f"{'is' if len(missing_products) == 1 else 'are'} "
                "not present in the ODC Database",
                file=sys.stderr,
            )
            sys.exit(1)

    is_glob = True
    # we assume the uri to be an absolute URL if it contains no wildcards
    # or if there are multiple uri values provided
    if (len(uris) > 1) or ("*" not in uris[0]):
        is_glob = False
        for url in uris:
            if "*" in url:
                logging.warning(
                    "A list of uris is assumed to include only absolute URLs. "
                    "Any wildcard characters will be escaped."
                )
    # Get a generator from supplied S3 Uri for candidate documents
    # Grab the URL from the resulting S3 item
    if is_glob:
        fetcher = S3Fetcher(aws_unsigned=no_sign_request)
        document_stream = fetcher(
            url.url
            for url in s3_find_glob(uris[0], skip_check=skip_check, s3=fetcher, **opts)
        )
    else:
        # if working with absolute URLs, no need for all the globbing logic
        document_stream = SimpleFetcher(
            aws_unsigned=no_sign_request, request_opts=opts
        )(uris)

    if url_string_replace:
        url_string_replace_tuple = tuple(url_string_replace.split(","))
        if len(url_string_replace_tuple) != 2:
            print(
                "url_string_replace must be two strings separated by a comma",
                file=sys.stderr,
            )
            sys.exit(1)
    else:
        url_string_replace_tuple = None

    doc2ds = Doc2Dataset(
        dc.index,
        products=candidate_products,
        skip_lineage=skip_lineage,
        fail_on_missing_lineage=fail_on_missing_lineage,
        verify_lineage=verify_lineage,
    )

    added = 0
    failed = 0
    skipped = 0
    found_docs = False
    try:
        for uri, dataset in parse_doc_stream(
            ((doc.url, doc.data) for doc in document_stream), on_error=doc_error
        ):
            if dataset is None:
                skipped += 1
                continue
            found_docs = True
            if convert_bools:
                for prop, val in dataset["properties"].items():
                    if val is True:
                        dataset["properties"][prop] = "true"
                    elif val is False:
                        dataset["properties"][prop] = "false"
            stac_doc = None
            if stac:
                item = Item.from_dict(dataset)
                if get_self_link(item) is None:
                    item.set_self_href(uri)
                dataset, new_uri, stac_doc = item_to_meta_uri(
                    item,
                    dc,
                    rename_product=rename_product,
                    url_string_replace=url_string_replace_tuple,
                )
                uri = new_uri or uri
            try:
                index_update_dataset(
                    dataset,
                    uri,
                    dc,
                    doc2ds,
                    update=update,
                    update_if_exists=update_if_exists,
                    allow_unsafe=allow_unsafe,
                    archive_less_mature=archive_less_mature,
                    publish_action=publish_action,
                    stac_doc=stac_doc,
                )
                added += 1
            except IndexingError:
                logging.exception("Failed to index dataset %s", uri)
                failed += 1
            except DatasetExists:
                skipped += 1
    except OSError as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)

    if not found_docs:
        print("Doc stream was empty", file=sys.stderr)
        sys.exit(1)

    print(
        f"Added {added} datasets, skipped {skipped} datasets and failed {failed} datasets."
    )
    if statsd_setting:
        statsd_gauge_reporting(added, ["app:s3_to_dc", "action:added"], statsd_setting)
        statsd_gauge_reporting(
            skipped, ["app:s3_to_dc", "action:skipped"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:s3_to_dc", "action:failed"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



"""
Test for stac_transform
"""

from datacube.utils.changes import get_doc_changes
from odc.apps.dc_tools._stac import stac_transform


def test_esri_lulc_stac_transform(esri_lulc_stac) -> None:
    transformed_stac_doc = stac_transform(esri_lulc_stac)
    expected_geometry_coordinates = (
        (
            (661287.0, 6209183.0),
            (614583.0, 6232873.0),
            (613410.0, 6233822.0),
            (612362.0, 6235174.0),
            (578054.0, 6285693.0),
            (577326.0, 6286947.0),
            (516925.0, 6401201.0),
            (515881.0, 6403393.0),
            (515178.0, 6406016.0),
            (514941.0, 6407721.0),
            (514850.0, 6409474.0),
            (514909.0, 6411230.0),
            (563156.0, 6887656.0),
            (563528.0, 6890381.0),
            (563901.0, 6891859.0),
            (564375.0, 6893191.0),
            (565363.0, 6895006.0),
            (566609.0, 6896326.0),
            (590299.0, 6919805.0),
            (591572.0, 6920477.0),
            (612179.0, 6928653.0),
            (624494.0, 6931468.0),
            (626035.0, 6931619.0),
            (627092.0, 6931319.0),
            (639943.0, 6927586.0),
            (643145.0, 6926090.0),
            (643924.0, 6925405.0),
            (644712.0, 6924404.0),
            (645547.0, 6922879.0),
            (646233.0, 6921043.0),
            (646681.0, 6919325.0),
            (647084.0, 6916926.0),
            (671635.0, 6559162.0),
            (671926.0, 6552687.0),
            (678860.0, 6397314.0),
            (687071.0, 6210141.0),
            (661287.0, 6209183.0),
        ),
    )

    assert (
        transformed_stac_doc["geometry"]["coordinates"] == expected_geometry_coordinates
    )


def test_landsat_stac_transform(landsat_stac, landsat_odc) -> None:
    transformed_stac_doc = stac_transform(landsat_stac)
    assert len(get_doc_changes(transformed_stac_doc, landsat_odc)) == 5


def test_sentinel_stac_transform(sentinel_stac_old, sentinel_odc) -> None:
    transformed_stac_doc = stac_transform(sentinel_stac_old)
    doc_changes = get_doc_changes(transformed_stac_doc, sentinel_odc)
    assert len(doc_changes) == 1


def test_sentinel_c1_stac_transform(sentinel_c1_stac, sentinel_c1_odc) -> None:
    transformed_stac_doc = stac_transform(sentinel_c1_stac)
    doc_changes = get_doc_changes(transformed_stac_doc, sentinel_c1_odc)
    assert len(doc_changes) == 0
    # test that absolute links are maintained for accessories
    assert (
        transformed_stac_doc["accessories"]["thumbnail"]["path"]
        == sentinel_c1_stac["assets"]["thumbnail"]["href"]
    )
    assert (
        transformed_stac_doc["accessories"]["product_metadata"]["path"]
        == sentinel_c1_stac["assets"]["product_metadata"]["href"]
    )
    assert (
        transformed_stac_doc["accessories"]["tileinfo_metadata"]["path"]
        == sentinel_c1_stac["assets"]["tileinfo_metadata"]["href"]
    )
    assert (
        transformed_stac_doc["accessories"]["granule_metadata"]["path"]
        == sentinel_c1_stac["assets"]["granule_metadata"]["href"]
    )


def test_sentinel_c1_rel_stac_transform(
    sentinel_c1_rel_stac, sentinel_c1_rel_odc
) -> None:
    transformed_stac_doc = stac_transform(sentinel_c1_rel_stac)
    doc_changes = get_doc_changes(transformed_stac_doc, sentinel_c1_rel_odc)
    assert len(doc_changes) == 0
    # test that absolute links are converted to relative links for accessories
    assert transformed_stac_doc["accessories"]["thumbnail"]["path"] == "L2A_PVI.jpg"
    assert (
        transformed_stac_doc["accessories"]["product_metadata"]["path"]
        == "product_metadata.xml"
    )
    assert (
        transformed_stac_doc["accessories"]["tileinfo_metadata"]["path"]
        == "tileInfo.json"
    )
    assert (
        transformed_stac_doc["accessories"]["granule_metadata"]["path"]
        == "metadata.xml"
    )


def test_usgs_landsat_stac_transform(usgs_landsat_stac) -> None:
    transformed_stac_doc = stac_transform(usgs_landsat_stac)

    expected_geometry_coordinates = (
        (
            (1087485.0, -313215.0),
            (904635.0, -356445.0),
            (860745.0, -172485.0),
            (1043085.0, -128325.0),
            (1087485.0, -313215.0),
        ),
    )
    assert (
        transformed_stac_doc["geometry"]["coordinates"] == expected_geometry_coordinates
    )


def test_stac_versions(landsat_stac, stac_1_1) -> None:
    transformed_stac_doc_1_0 = stac_transform(landsat_stac)
    transformed_stac_doc_1_1 = stac_transform(stac_1_1)
    assert transformed_stac_doc_1_0["crs"] == "EPSG:32656"
    assert transformed_stac_doc_1_1["crs"] == "EPSG:32656"


def test_lidar_stac_transform(lidar_stac) -> None:
    transformed_stac_doc = stac_transform(lidar_stac)
    expected_geometry_coordinates = (
        (
            (766000.0, 6732000.0),
            (766000.0, 6730000.0),
            (768000.0, 6730000.0),
            (768000.0, 6732000.0),
            (766000.0, 6732000.0),
        ),
    )
    assert (
        transformed_stac_doc["geometry"]["coordinates"] == expected_geometry_coordinates
    )


def test_world_wrapping_stac(world_wrapping_stac) -> None:
    transformed_stac_doc = stac_transform(world_wrapping_stac)
    expected_geometry_coordinates = (
        (3336000.0, -2079990.0),
        (3336000.0, -1983990.0),
        (3432030.0, -1983990.0),
        (3432030.0, -2079990.0),
        (3336000.0, -2079990.0),
    )
    assert (
        transformed_stac_doc["geometry"]["coordinates"][0]
        == expected_geometry_coordinates
    )



#!/usr/bin/env python3
"""
Index the Copernicus DEM automatically.
"""
import click
import concurrent.futures
import logging
import pystac
import rasterio
import sys
from math import ceil, floor
from rio_stac import create_stac_item
from typing import Tuple

from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.metadata import stac2ds
from datacube.ui.click import environment_option, pass_config
from datacube.utils import read_documents
from odc.apps.dc_tools.utils import (
    DatasetExists,
    archive_less_mature,
    bbox,
    index_update_dataset,
    limit,
    update_if_exists_flag,
    publish_action,
    statsd_gauge_reporting,
    statsd_setting,
)
from sqlalchemy.exc import OperationalError, ProgrammingError

PRODUCTS = {
    "cop_30": (
        "https://raw.githubusercontent.com/opendatacube/"
        "datacube-dataset-config/master/products/dem_cop_30.odc-product.yaml"
    ),
    "cop_90": (
        "https://raw.githubusercontent.com/opendatacube/"
        "datacube-dataset-config/master/products/dem_cop_90.odc-product.yaml"
    ),
}

# URIs need north/south, which is N00 and east/west, which is E000
URI_TEMPLATES = {
    "cop_30": (
        "https://copernicus-dem-30m.s3.eu-central-1.amazonaws.com/"
        "Copernicus_DSM_COG_10_{ns}_00_{ew}_00_DEM/Copernicus_DSM_COG_10_{ns}_00_{ew}_00_DEM.tif"
    ),
    "cop_90": (
        "https://copernicus-dem-90m.s3.eu-central-1.amazonaws.com/"
        "Copernicus_DSM_COG_30_{ns}_00_{ew}_00_DEM/Copernicus_DSM_COG_30_{ns}_00_{ew}_00_DEM.tif"
    ),
}


def add_cop_dem_product(dc: Datacube, product):
    if product in PRODUCTS:
        product_uri = PRODUCTS[product]
    else:
        raise ValueError(f"Unknown product {product}")

    for _, doc in read_documents(product_uri):
        dc.index.products.add_document(doc)
    print(f"Product definition added for {product}")


def get_dem_tile_uris(bounding_box, product):
    # Validate the bounding_box
    if bounding_box is None:
        logging.warning(
            "No BBOX provided, running full extent... this will take a long time."
        )
        bounding_box = (-180, -90, 180, 90)
    else:
        bounding_box = bounding_box.split(",")
        if len(bounding_box) != 4:
            raise ValueError("bounding_box must be in the format: minx,miny,maxx,maxy")
        bounding_box = [float(x) for x in bounding_box]

    # Get the uris
    left = bounding_box[0]
    right = bounding_box[2]
    bottom = bounding_box[1]
    top = bounding_box[3]

    x_range = range(floor(left), ceil(right))
    y_range = range(floor(bottom), ceil(top))

    for x in x_range:
        for y in y_range:
            if x < 0:
                x_str = f"W{abs(x):03d}"
            else:
                x_str = f"E{x:03d}"
            if y < 0:
                y_str = f"S{abs(y):02d}"
            else:
                y_str = f"N{y:02d}"
            yield (
                URI_TEMPLATES[product].format(ns=y_str, ew=x_str),
                f"{x_str}_{y_str}",
            )


def process_uri_tile(
    uri_tile: Tuple[str, str, str],
    product: str,
    dc: Datacube,
    doc2ds: Doc2Dataset,
    update_if_exists: bool = True,
    archive_less_mature: int | None = None,
    publish_action: str | None = None,
) -> Tuple[pystac.Item, str]:
    product_name = f"dem_{product}"
    uri, tile = uri_tile
    properties = {
        "odc:product": product_name,
        "odc:region_code": tile,
        "start_datetime": "1900-01-01",
        "end_datetime": "2100-01-01",
    }

    with rasterio.Env(aws_unsigned=True, GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR"):
        item = create_stac_item(
            uri,
            collection=product_name,
            with_proj=True,
            properties=properties,
            asset_media_type=pystac.MediaType.COG,
            asset_name="elevation",
        )

    index_update_dataset(
        next(stac2ds([item])),
        uri,
        dc,
        doc2ds,
        update_if_exists=update_if_exists,
        allow_unsafe=True,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        stac_doc=item.to_dict(),
    )

    return True


def cop_dem_to_dc(
    dc: Datacube,
    product: str,
    bounding_box,
    limit: int,
    update: bool,
    n_workers: int = 100,
    archive_less_mature=None,
    publish_action=None,
) -> Tuple[int, int]:
    doc2ds = Doc2Dataset(dc.index)

    # Get a generator of (uris)
    uris_tiles = list(get_dem_tile_uris(bounding_box, product))
    if limit:
        uris_tiles = uris_tiles[0:limit]

    # Do the indexing of all the things
    success = 0
    failure = 0
    skipped = 0

    sys.stdout.write(f"Starting Cop DEM indexing with {n_workers} workers...\n")

    with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
        future_to_uri = {
            executor.submit(
                process_uri_tile,
                uri_tile,
                product,
                dc,
                doc2ds,
                update_if_exists=update,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
            ): uri_tile[0]
            for uri_tile in uris_tiles
        }
        for future in concurrent.futures.as_completed(future_to_uri):
            uri = future_to_uri[future]
            try:
                _ = future.result()
                success += 1
                if success % 10 == 0:
                    sys.stdout.write(f"\rAdded {success} datasets...")
            except DatasetExists:
                skipped += 1
            except rasterio.errors.RasterioIOError:
                logging.info("Couldn't read file %s", uri, exc_info=True)
            except Exception:  # pylint:disable=broad-except
                logging.exception("Failed to handle uri %s", uri)
                failure += 1
    sys.stdout.write("\r")

    return success, failure, skipped


@click.command("cop-dem-to-dc")
@environment_option
@pass_config
@limit
@update_if_exists_flag
@bbox
@statsd_setting
@archive_less_mature
@publish_action
@click.option(
    "--product",
    default="cop_30",
    help="Product to add to the index, either cop_30 or cop_90",
)
@click.option(
    "--add-product",
    is_flag=True,
    default=False,
    help="If set, add the product too",
)
@click.option(
    "--workers",
    default=20,
    type=int,
    help="Number of threads to use to process, default 20",
)
def cli(
    cfg_env,
    limit,
    update_if_exists,
    bbox,
    statsd_setting,
    product,
    add_product,
    workers,
    archive_less_mature,
    publish_action,
):
    """
    Index the Copernicus DEM automatically.
    """
    if product not in PRODUCTS:
        raise ValueError(
            f"Unknown product {product}, must be one of {' '.join(PRODUCTS)}"
        )
    try:
        dc = Datacube(env=cfg_env, app="cop-dem-to-dc")
    except (OperationalError, ProgrammingError) as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)

    if add_product:
        add_cop_dem_product(dc, product)

    print(f"Indexing Copernicus DEM for {product} with bounding box of {bbox}")

    added, failed, skipped = cop_dem_to_dc(
        dc,
        product,
        bbox,
        limit,
        update_if_exists,
        n_workers=workers,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
    )

    print(
        f"Added {added} Datasets, failed {failed} Datasets, skipped {skipped} Datasets"
    )

    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:cop_dem_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:cop_dem_to_dc", "action:failed"], statsd_setting
        )
        statsd_gauge_reporting(
            skipped, ["app:cop_dem_to_dc", "action:skipped"], statsd_setting
        )

    if failed > 0:
        sys.exit(failed)


if __name__ == "__main__":
    cli()



# Tests using the Click framework the s3-to-dc CLI tool
# flake8: noqa
import boto3
import json
import pytest
from moto import mock_aws
from click.testing import CliRunner
from odc.aws.queue import get_queue, get_queues, redrive_queue
from odc.apps.dc_tools.s3_to_dc import cli as s3_to_dc
from odc.apps.dc_tools.redrive_to_queue import cli as redrive_cli


ALIVE_QUEUE_NAME = "mock-alive-queue"
DEAD_QUEUE_NAME = "mock-dead-queue"


@mock_aws
def test_redrive_to_queue_cli(aws_env) -> None:
    resource = boto3.resource("sqs")

    dead_queue = resource.create_queue(QueueName=DEAD_QUEUE_NAME)
    resource.create_queue(
        QueueName=ALIVE_QUEUE_NAME,
        Attributes={
            "RedrivePolicy": json.dumps(
                {
                    "deadLetterTargetArn": dead_queue.attributes.get("QueueArn"),
                    "maxReceiveCount": 2,
                }
            ),
        },
    )

    for i in range(35):
        dead_queue.send_message(MessageBody=json.dumps({"content": f"Something {i}"}))
    runner = CliRunner()
    # Invalid value string
    returned = runner.invoke(
        redrive_cli,
        [str(DEAD_QUEUE_NAME), str(ALIVE_QUEUE_NAME), "--limit", "string_test"],
    )
    assert returned.exit_code == 1, f"Output: {returned.output}"

    # Invalid value 0
    returned = runner.invoke(
        redrive_cli,
        [str(DEAD_QUEUE_NAME), str(ALIVE_QUEUE_NAME), "--limit", 0],
    )
    assert returned.exit_code == 1, f"Output: {returned.output}"

    # Valid value 1
    returned = runner.invoke(
        redrive_cli,
        [str(DEAD_QUEUE_NAME), str(ALIVE_QUEUE_NAME), "--limit", 1],
    )
    assert returned.exit_code == 0, f"Output: {returned.output}"
    assert (
        int(get_queue(ALIVE_QUEUE_NAME).attributes.get("ApproximateNumberOfMessages"))
        == 1
    )

    # Valid value None (all)
    returned = runner.invoke(
        redrive_cli,
        [str(DEAD_QUEUE_NAME), str(ALIVE_QUEUE_NAME), "--limit", None],
    )
    assert returned.exit_code == 0, f"Output: {returned.output}"
    assert (
        int(get_queue(DEAD_QUEUE_NAME).attributes.get("ApproximateNumberOfMessages"))
        == 0
    )


def test_s3_to_dc_skips_already_indexed_datasets(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    runner = CliRunner()
    # This will fail if requester pays is enabled
    results = [
        runner.invoke(
            s3_to_dc,
            [
                "--no-sign-request",
                "s3://odc-tools-test/cemp_insar/**/*.yaml",
                "cemp_insar_alos_displacement",
                "--env",
                env_name,
            ],
        )
        for _ in range(1, 3)
    ]

    # The first run should succeed and index all 25 datasets
    assert results[0].exit_code == 0, f"Output: {results[0].output}"
    assert (
        results[0].output
        == "Added 25 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )

    # The second run should succeed by SKIPPING all 25 datasets
    assert results[1].exit_code == 0, f"Output: {results[1].output}"
    assert (
        results[1].output
        == "Added 0 datasets, skipped 25 datasets and failed 0 datasets.\n"
    )


def test_s3_to_dc_stac(
    mocked_s3_datasets, aws_env, odc_test_db_with_products, env_name
) -> None:
    result = CliRunner().invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "--stac",
            "s3://odc-tools-test/sentinel-s2-l2a-cogs/31/Q/GB/2020/8/S2B_31QGB_20200831_0_L2A/*_L2A.json",
            "--rename-product",
            "s2_l2a",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )


def test_s3_to_dc_stac_update_if_exist(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    result = CliRunner().invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "--stac",
            "--update-if-exists",
            "s3://odc-tools-test/sentinel-s2-l2a-cogs/31/Q/GB/2020/8/S2B_31QGB_20200831_0_L2A/*_L2A.json",
            "--rename-product",
            "s2_l2a",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )


def test_s3_to_dc_stac_update_if_exist_allow_unsafe(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    runner = CliRunner()
    result = runner.invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "--stac",
            "--update-if-exists",
            "--allow-unsafe",
            "s3://odc-tools-test/sentinel-s2-l2a-cogs/31/Q/GB/2020/8/S2B_31QGB_20200831_0_L2A/*_L2A.json",
            "--rename-product",
            "s2_l2a",
            "--env",
            env_name,
        ],
    )
    print(f"s3-to-dc exit_code: {result.exit_code}, output:{result.output}")
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )


def test_s3_to_dc_fails_to_index_non_dataset_yaml(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    runner = CliRunner()
    result = runner.invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "s3://odc-tools-test/baseline/ga_s2am_ard_3/49/JFM/2016/12/14/20161214T092514/ga_s2am_ard_3-2-1_49JFM_2016-12-14_final.odc-metadata.yaml",
            "ga_ls5t_nbart_gm_cyear_3",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 1, f"Output: {result.output}"
    assert (
        result.output == "Added 0 datasets, skipped 0 datasets and failed 1 datasets.\n"
    )


def test_s3_to_dc_partially_succeeds_when_given_invalid_and_valid_dataset_yamls(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    runner = CliRunner()
    result = runner.invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "--skip-lineage",
            # This folder contains two yaml one valid dataset yaml and another non dataset yaml
            "s3://odc-tools-test/derivative/ga_ls5t_nbart_gm_cyear_3/3-0-0/x08/y23/1994--P1Y/*.yaml",
            "ga_ls5t_nbart_gm_cyear_3",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 1, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 1 datasets.\n"
    )


def test_s3_to_dc_list_absolute_urls(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    # provide mulitple uris, as absolute URLs
    runner = CliRunner()
    result = runner.invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "s3://odc-tools-test/cemp_insar/01/07/alos_cumul_2010-01-07.yaml",
            "s3://odc-tools-test/cemp_insar/04/01/alos_cumul_2010-04-01.yaml",
            "s3://odc-tools-test/cemp_insar/08/11/alos_cumul_2010-08-11.yaml",
            "cemp_insar_alos_displacement",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 3 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )


def test_s3_to_dc_no_product(
    mocked_s3_datasets, odc_test_db_with_products, env_name
) -> None:
    # product should not need to be specified
    runner = CliRunner()
    result = runner.invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "s3://odc-tools-test/cemp_insar/01/07/alos_cumul_2010-01-07.yaml",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )

    # test with glob
    result2 = CliRunner().invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "--stac",
            "s3://odc-tools-test/sentinel-s2-l2a-cogs/31/Q/GB/2020/8/S2B_31QGB_20200831_0_L2A/*_L2A.json",
            "--rename-product",
            "s2_l2a",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result2.exit_code == 0, f"Output: {result2.output}"
    assert (
        result2.output
        == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )


def test_convert_bools(mocked_s3_datasets, odc_test_db_with_products, env_name) -> None:
    dc = odc_test_db_with_products
    assert dc.index.products.get_by_name("ga_s1_iw_hh_c1") is not None
    runner = CliRunner()
    result = runner.invoke(
        s3_to_dc,
        [
            "--no-sign-request",
            "--convert-bools",
            "--stac",
            "s3://odc-tools-test/experimental/linkage/s1_rtc_c1/t007_014549_iw1/2025/1/29/*.json",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    ds = dc.index.datasets.get("5f94dd81-241f-559a-9362-05b223d45ae1")
    # boolean values converted to strings
    assert ds.metadata.noise_removal_applied == "true"
    # other values left as they are
    assert ds.metadata.speckle_filter_applied == "False"
    assert ds.metadata_doc["properties"]["sarard:speckle_filter_window"] == []



import boto3
import json
import os
import pytest
from click.testing import CliRunner
from moto import mock_aws
from pathlib import Path

from odc.apps.dc_tools.fs_to_dc import cli as fs_cli
from odc.apps.dc_tools.s3_to_dc import cli as s3_cli
from odc.apps.dc_tools.sqs_to_dc import cli as sqs_cli


@pytest.fixture
def aws_credentials() -> None:
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"


@pytest.fixture()
def sns_setup(aws_credentials, aws_env):
    """Set up SNS topic and SQS queue subscribed to it

    Tests are structured as follows:
    input: [ STAC -> SNS -> SQS ] -> dc_tools -> output: [ STAC -> SNS -> SQS ]
    """
    with mock_aws():
        sns = boto3.client("sns")
        sqs = boto3.client("sqs")

        # Set up input topic and queue which dc_tools will receive from
        input_topic = sns.create_topic(Name="input-topic")
        input_topic_arn = input_topic.get("TopicArn")

        input_queue_name = "input-queue"
        input_queue = sqs.create_queue(QueueName=input_queue_name)
        input_attrs = sqs.get_queue_attributes(
            QueueUrl=input_queue.get("QueueUrl"), AttributeNames=["QueueArn"]
        )
        input_queue_arn = input_attrs["Attributes"]["QueueArn"]

        sns.subscribe(
            TopicArn=input_topic_arn,
            Protocol="sqs",
            Endpoint=input_queue_arn,
        )

        # Set up output topic and queue, which dc_tools will publish results to
        output_topic = sns.create_topic(Name="output-topic")
        output_topic_arn = output_topic.get("TopicArn")

        output_queue_name = "test-queue"
        output_queue = sqs.create_queue(QueueName=output_queue_name)
        output_attrs = sqs.get_queue_attributes(
            QueueUrl=output_queue.get("QueueUrl"), AttributeNames=["QueueArn"]
        )
        output_queue_arn = output_attrs["Attributes"]["QueueArn"]

        sns.subscribe(
            TopicArn=output_topic_arn,
            Protocol="sqs",
            Endpoint=output_queue_arn,
        )

        yield sns, input_topic_arn, output_topic_arn, sqs, input_queue_name, output_queue.get(
            "QueueUrl"
        )


def test_s3_publishing_action_from_stac(
    mocked_s3_datasets, odc_test_db_with_products, env_name, s2am_dsid, sns_setup
) -> None:
    _, _, output_topic_arn, sqs, _, output_queue_url = sns_setup

    dc = odc_test_db_with_products
    assert dc.index.datasets.get(s2am_dsid) is None

    runner = CliRunner()
    # This will fail if requester pays is enabled
    result = runner.invoke(
        s3_cli,
        [
            "--no-sign-request",
            "--stac",
            "--skip-lineage",
            "--update-if-exists",
            f"--publish-action={output_topic_arn}",
            "s3://odc-tools-test/baseline/ga_s2am_ard_3/49/JFM/2016/12/14/20161214T092514/*stac-item.json",
            "ga_s2am_ard_3",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )

    print(f"s3-to-dc exit_code: {result.exit_code}, output:{result.output}")
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )
    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
    )["Messages"]
    assert len(messages) == 1
    message_attrs = json.loads(messages[0]["Body"]).get("MessageAttributes")
    assert message_attrs["action"].get("Value") == "ADDED"
    assert message_attrs["product"].get("Value") == "ga_s2am_ard_3"


def test_s3_publishing_action_from_eo3(
    mocked_s3_datasets, odc_test_db_with_products, env_name, s2am_dsid, sns_setup
) -> None:
    """Same as above but requiring stac to eo3 conversion"""
    _, _, output_topic_arn, sqs, _, output_queue_url = sns_setup

    dc = odc_test_db_with_products
    assert dc.index.datasets.get(s2am_dsid) is None

    runner = CliRunner()
    result = runner.invoke(
        s3_cli,
        [
            "--no-sign-request",
            "--update-if-exists",
            "--skip-lineage",
            f"--publish-action={output_topic_arn}",
            "s3://odc-tools-test/baseline/ga_s2am_ard_3/49/JFM/2016/12/14/20161214T092514/*odc-metadata.yaml",
            "ga_s2am_ard_3",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )

    print(f"s3-to-dc exit_code: {result.exit_code}, output:{result.output}")
    assert result.exit_code == 0, f"Output: {result.output}"
    assert (
        result.output == "Added 1 datasets, skipped 0 datasets and failed 0 datasets.\n"
    )

    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
    )["Messages"]
    assert len(messages) == 1
    message_attrs = json.loads(messages[0]["Body"]).get("MessageAttributes")
    assert message_attrs["action"].get("Value") == "ADDED"


TEST_DATA_FOLDER: Path = Path(__file__).parent.joinpath("data")
STAC_DATA: str = "ga_ls5t_nbart_gm_cyear_3_x30y14_1999--P1Y_final.stac-item.json"


@pytest.fixture
def stac_doc():
    with TEST_DATA_FOLDER.joinpath(STAC_DATA).open("r") as f:
        return f.read()


def test_sqs_publishing(
    aws_credentials, aws_env, stac_doc, odc_test_db_with_products, env_name, sns_setup
) -> None:
    """Test that actions are published with sqs_to_dc"""
    (
        _,
        input_topic_arn,
        output_topic_arn,
        sqs,
        input_queue_name,
        output_queue_url,
    ) = sns_setup

    input_queue = sqs.create_queue(QueueName=input_queue_name)
    sqs.send_message(
        QueueUrl=input_queue.get("QueueUrl"),
        MessageBody=json.dumps(
            {
                "Type": "Notification",
                "MessageId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxxxxxxxxx",
                "TopicArn": input_topic_arn,
                "Message": stac_doc,
                "MessageAttributes": {
                    "action": {"DataType": "String", "StringValue": "ARCHIVED"}
                },
            }
        ),
    )

    runner = CliRunner()
    result = runner.invoke(
        sqs_cli,
        [
            input_queue_name,
            "ga_ls5t_nbart_gm_cyear_3",
            "--skip-lineage",
            "--no-sign-request",
            "--update-if-exists",
            "--stac",
            f"--publish-action={output_topic_arn}",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    print(f"sqs-to-dc exit_code: {result.exit_code}, output:{result.output}")

    assert result.exit_code == 0, f"Output: {result.output}"

    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
    )["Messages"]
    assert len(messages) == 1
    message_attrs = json.loads(messages[0]["Body"]).get("MessageAttributes")
    assert message_attrs["action"].get("Value") == "ADDED"
    assert message_attrs["product"].get("Value") == "ga_ls5t_nbart_gm_cyear_3"
    assert message_attrs["maturity"].get("Value") == "final"


def test_sqs_publishing_archive_flag(
    aws_credentials,
    aws_env,
    stac_doc,
    odc_db_for_archive,
    env_name,
    ls5t_dsid,
    sns_setup,
) -> None:
    """Test that an ARCHIVE SNS message is published when the --archive flag is used."""
    (
        sns,
        input_topic_arn,
        output_topic_arn,
        sqs,
        input_queue_name,
        output_queue_url,
    ) = sns_setup

    sns.publish(
        TopicArn=input_topic_arn,
        Message=stac_doc,
        MessageAttributes={"action": {"DataType": "String", "StringValue": "ARCHIVED"}},
    )

    dc = odc_db_for_archive
    assert dc.index.datasets.get(ls5t_dsid) is not None

    runner = CliRunner()
    result = runner.invoke(
        sqs_cli,
        [
            input_queue_name,
            "ga_ls5t_nbart_gm_cyear_3",
            "--skip-lineage",
            "--no-sign-request",
            "--update-if-exists",
            "--stac",
            "--archive",
            f"--publish-action={output_topic_arn}",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    print(f"sqs-to-dc exit_code: {result.exit_code}, output:{result.output}")

    assert result.exit_code == 0, f"Output: {result.output}"

    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
    )["Messages"]
    assert len(messages) == 1
    message_attrs = json.loads(messages[0]["Body"]).get("MessageAttributes")
    assert message_attrs["action"].get("Value") == "ARCHIVED"
    assert dc.index.datasets.get(ls5t_dsid).is_archived is True


def test_sqs_publishing_archive_attribute(
    aws_credentials,
    aws_env,
    stac_doc,
    odc_db_for_archive,
    env_name,
    ls5t_dsid,
    sns_setup,
) -> None:
    """Test that archiving occurs when ARCHIVED is in the message attributes"""
    (
        sns,
        input_topic_arn,
        output_topic_arn,
        sqs,
        input_queue_name,
        output_queue_url,
    ) = sns_setup

    sns.publish(
        TopicArn=input_topic_arn,
        Message=stac_doc,
        MessageAttributes={"action": {"DataType": "String", "StringValue": "ARCHIVED"}},
    )

    dc = odc_db_for_archive
    ds = dc.index.datasets.get(ls5t_dsid)
    assert ds is not None
    assert ds.is_archived is False

    runner = CliRunner()
    result = runner.invoke(
        sqs_cli,
        [
            input_queue_name,
            "ga_ls5t_nbart_gm_cyear_3",
            "--skip-lineage",
            "--no-sign-request",
            "--update-if-exists",
            "--stac",
            f"--publish-action={output_topic_arn}",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )

    assert result.exit_code == 0, f"Output: {result.output}"
    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
        MessageAttributeNames=["All"],
    )["Messages"]
    assert len(messages) == 1
    message_attrs = json.loads(messages[0]["Body"]).get("MessageAttributes")
    assert message_attrs["action"].get("Value") == "ARCHIVED"
    assert dc.index.datasets.get(ls5t_dsid).is_archived is True


def test_with_archive_less_mature(
    aws_credentials,
    aws_env,
    odc_db,
    env_name,
    nrt_dsid,
    final_dsid,
    sns_setup,
) -> None:
    _, _, output_topic_arn, sqs, _, output_queue_url = sns_setup

    dc = odc_db
    assert dc.index.datasets.get(nrt_dsid) is None

    runner = CliRunner()
    nrt_result = runner.invoke(
        fs_cli,
        [
            str(TEST_DATA_FOLDER),
            "--glob=**/maturity-nrt.odc-metadata.yaml",
            "--archive-less-mature",
            f"--publish-action={output_topic_arn}",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    print(f"fs-to-dc exit_code: {nrt_result.exit_code}, " "output:{nrt_result.output}")

    assert nrt_result.exit_code == 0, f"Output: {nrt_result.output}"
    assert dc.index.datasets.get(nrt_dsid) is not None

    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
    )["Messages"]
    assert len(messages) == 1
    message = messages[0]
    message_attrs = json.loads(message["Body"]).get("MessageAttributes")
    assert message_attrs["action"].get("Value") == "ADDED"
    body = json.loads(message["Body"])
    assert json.loads(body["Message"]).get("id") == nrt_dsid

    assert dc.index.datasets.get(final_dsid) is None

    final_result = runner.invoke(
        fs_cli,
        [
            str(TEST_DATA_FOLDER),
            "--glob=**/maturity-final.odc-metadata.yaml",
            "--archive-less-mature",
            f"--publish-action={output_topic_arn}",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    print(
        f"fs-to-dc exit_code: {final_result.exit_code}, " "output:{final_result.output}"
    )

    assert final_result.exit_code == 0, f"Output: {final_result.output}"
    assert dc.index.datasets.get(final_dsid) is not None

    messages = sqs.receive_message(
        QueueUrl=output_queue_url,
        MaxNumberOfMessages=10,
    )["Messages"]
    assert len(messages) == 2

    nrt_message = messages[0]
    nrt_attrs = json.loads(nrt_message["Body"]).get("MessageAttributes")
    assert nrt_attrs["action"].get("Value") == "ARCHIVED"

    nrt_body = json.loads(nrt_message["Body"])
    assert json.loads(nrt_body["Message"]).get("id") == nrt_dsid

    final_message = messages[1]
    final_attrs = json.loads(final_message["Body"]).get("MessageAttributes")
    assert final_attrs["action"].get("Value") == "ADDED"
    final_body = json.loads(final_message["Body"])
    assert json.loads(final_body["Message"]).get("id") == final_dsid



import click
import logging
import signal
import sys
import tarfile
from odc.aio import S3Fetcher
from odc.io import read_stdin_lines
from odc.io.tar import add_txt_file, tar_mode
from odc.io.timer import RateEstimator
from sys import stderr, stdout


@click.command("s3-to-tar")
@click.option("-n", type=int, help="Number of concurrent async connections to S3")
@click.option("--verbose", "-v", is_flag=True, help="Be verbose")
@click.option("--gzip", is_flag=True, help="Compress with gzip")
@click.option("--xz", is_flag=True, help="Compress with xz")
@click.option("--no-sign-request", is_flag=True, help="Do not sign AWS S3 requests")
@click.option(
    "--request-payer",
    is_flag=True,
    help="Needed when accessing requester pays public buckets",
)
@click.argument("outfile", type=str, nargs=1, default="-")
def cli(
    n, verbose, gzip, xz, outfile, no_sign_request=None, request_payer=False
) -> None:
    """Fetch a bunch of s3 files into a tar archive.

    \b
    For every non-empty line in stdin
       - Treat line as a URI and fetch document from it
       - Write content of the file to a tar archive using `bucket-name/path/to/file` as file name
    """

    opts = {}
    if request_payer:
        opts["RequestPayer"] = "requester"

    logging.basicConfig(
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
        level=logging.ERROR,
    )

    nconnections = 24 if n is None else n
    exit_early = False

    def dump_to_tar(data_stream, tar) -> None:
        nonlocal exit_early  # noqa: F824
        fps = RateEstimator()

        for d in data_stream:
            fps()
            fname = d.url[5:]

            if d.data is not None:
                if verbose:
                    if fps.every(10):
                        print(".", file=stderr, end="", flush=True)

                    if fps.every(100):
                        print(" {}".format(str(fps)), file=stderr)

                add_txt_file(tar, fname, d.data, last_modified=d.last_modified)
            else:
                print("Failed %s (%s)" % (d.url, str(d.error)), file=stderr)

            if exit_early:
                break

        if verbose:
            print(" {}".format(str(fps)), file=stderr)

    fetcher = S3Fetcher(nconcurrent=nconnections, aws_unsigned=no_sign_request)
    is_pipe = outfile == "-"
    tar_opts = {"mode": "w" + tar_mode(gzip=gzip, xz=xz, is_pipe=is_pipe)}
    if is_pipe:
        if stdout.isatty():
            click.echo("Will not write to a terminal", err=True)
            sys.exit(1)
        # TODO: on windows switch stdout to binary mode
        tar_opts["fileobj"] = stdout.buffer
    else:
        tar_opts["name"] = outfile

    urls = read_stdin_lines(skip_empty=True)

    def on_ctrlc(sig, frame) -> None:
        nonlocal exit_early
        print("Shutting down...", file=sys.stderr)
        exit_early = True

    signal.signal(signal.SIGINT, on_ctrlc)

    with tarfile.open(**tar_opts) as tar:
        dump_to_tar(fetcher(urls, **opts), tar)

    fetcher.close()


if __name__ == "__main__":
    cli()



from pathlib import Path

import pytest
from click.testing import CliRunner

from odc.apps.dc_tools.esa_worldcover_to_dc import (
    _unpack_bbox,
    cli,
    get_tile_uris,
    URI_TEMPLATE,
)


@pytest.fixture
def bbox() -> str:
    return "5,5,7,7"


@pytest.fixture
def bbox_africa() -> str:
    return "-26.359944882003788,-47.96476498374171,64.4936701740102,38.34459242512347"


def test_bboxes() -> None:
    bbox = "0,0,1,1"
    bounding_box = [float(x) for x in bbox.split(",")]
    assert _unpack_bbox(bounding_box) == (0, 0, 3, 3)
    assert len(list(get_tile_uris(bbox))) == 1

    bbox = "0,0,3,3"
    bounding_box = [float(x) for x in bbox.split(",")]
    assert _unpack_bbox(bounding_box) == (0, 0, 3, 3)
    assert len(list(get_tile_uris(bbox))) == 1

    bbox = "1,1,5,5"
    bounding_box = [float(x) for x in bbox.split(",")]
    assert _unpack_bbox(bounding_box) == (0, 0, 6, 6)

    bbox = "-98,15,-90,30"
    bounding_box = [float(x) for x in bbox.split(",")]
    assert _unpack_bbox(bounding_box) == (-99, 15, -90, 30)
    assert len(list(get_tile_uris(bbox))) == 15


def test_get_dem_tile_uris(bbox) -> None:
    uris = list(get_tile_uris(bbox))

    assert uris[0][0] == (
        "https://esa-worldcover.s3.eu-central-1.amazonaws.com/"
        "v100/2020/map/ESA_WorldCover_10m_2020_v100_N03E003_Map.tif"
    )

    print(uris)
    assert len(uris) == 4


def test_complex_bbox(bbox_africa) -> None:
    uris = list(get_tile_uris(bbox_africa))

    assert len(uris) == 899


@pytest.fixture
def mock_esa_worldcover_datasets(monkeypatch) -> None:
    """Replace the fetching of remote ESA WorldCover datasets with local downsampled versions"""
    fname_template = URI_TEMPLATE.split("/")[-1]
    local_template = (
        "file://"
        + str(Path(__file__).parent.absolute())
        + f"/data/esa_worldcover/{fname_template}"
    )
    monkeypatch.setattr(
        "odc.apps.dc_tools.esa_worldcover_to_dc.URI_TEMPLATE", local_template
    )


def test_indexing_cli(
    bbox, odc_test_db_with_products, mock_esa_worldcover_datasets, env_name
) -> None:
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--bbox",
            bbox,
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"Output: {result.output}"



import click
import tarfile
from odc.azure import download_yamls, find_blobs
from odc.io.tar import add_txt_file, tar_mode


@click.command("azure-to-tar")
@click.argument("account_url", type=str, nargs=1)
@click.argument("container_name", type=str, nargs=1)
@click.argument("credential", type=str, nargs=1)
@click.argument("prefix", type=str, nargs=1)
@click.argument("suffix", type=str, nargs=1)
@click.option(
    "--workers", "-w", type=int, default=32, help="Number of threads to download blobs"
)
@click.option(
    "--outfile", type=str, default="metadata.tar.gz", help="Sets the output file name"
)
def cli(
    account_url: str,
    container_name: str,
    credential: str,
    prefix: str,
    suffix: str,
    workers: int,
    outfile: str,
) -> None:
    print(f"Opening AZ Container {container_name} on {account_url}")
    print(f"Searching on prefix '{prefix}' for files matching suffix '{suffix}'")
    yaml_urls = find_blobs(container_name, credential, prefix, suffix, account_url)

    print(f"Found {len(yaml_urls)} datasets")
    yamls = download_yamls(account_url, container_name, credential, yaml_urls, workers)

    url_prefix = (account_url + "/" + container_name + "/")[len("https://") :]

    # jam it all in a tar
    tar_opts = {
        "name": outfile,
        "mode": "w" + tar_mode(gzip=True, xz=True, is_pipe=False),
    }
    with tarfile.open(**tar_opts) as tar:
        for yaml in yamls:
            add_txt_file(tar=tar, content=yaml[0], fname=url_prefix + yaml[1])

    print("Done!")


if __name__ == "__main__":
    cli()



import click
import os
import shutil
import tarfile
from google.cloud import storage


@click.command("gs-to-tar")
@click.option(
    "--bucket", type=str, help="The Google Storage bucket, without gs:// prefix"
)
@click.option(
    "--prefix",
    type=str,
    help="The filepath to search for yaml files, i.e. geomedian/v2.1.0/L5",
)
@click.option(
    "--suffix", type=str, default=".yaml", help="The suffix of your metadata files"
)
@click.option(
    "--outfile", type=str, default="metadata.tar.gz", help="Sets the output file name"
)
def cli(bucket, prefix, suffix, outfile) -> None:
    """Download Metadata from GS bucket to tarball

    Example:

       \b
       Download files in directory that match `*yaml` and store them as a tar
        > gs-to-tar --bucket data.deadev.com --prefix mangrove_cover

    """

    # Connect to bucket
    client = storage.Client.create_anonymous_client()
    bucket_object = client.bucket(bucket, user_project=None)

    # Get list of files under prefix
    print(
        "Searching {bucket}/{prefix} for files".format(
            bucket=bucket_object, prefix=prefix
        )
    )
    blobs = bucket_object.list_blobs(prefix=prefix)

    # Filter on suffix
    files = [blob for blob in blobs if blob.name.endswith(suffix)]
    file_count = str(len(files))
    print("Found {0} metadata files".format(file_count))

    # Download Files to tar with an updating counter
    file_num = 1
    with tarfile.open(outfile, "w:gz") as tar:
        for yaml in files:
            count = str(file_num)
            filename = "{bucket}/{filepath}".format(bucket=bucket, filepath=yaml.name)

            # ensure dir exists
            if not os.path.exists(os.path.dirname("./" + filename)):
                os.makedirs(os.path.dirname("./" + filename))
            # download to tar
            yaml.download_to_filename(filename=filename, client=client)
            tar.add(filename)
            os.remove("./" + filename)

            # counter
            print(
                "{count}/{file_count} Downloaded".format(
                    count=count, file_count=file_count
                )
            )
            file_num += 1

    # Deletes the directory recursively
    shutil.rmtree("./" + bucket)

    print("Done!")


if __name__ == "__main__":
    cli()



import click
import re
import sys
from fnmatch import fnmatch
from datacube.utils.aws import s3_client
from odc.aws.inventory import list_inventory


def build_predicate(glob=None, regex=None, prefix=None):
    def match_prefix(entry):
        return entry.Key.startswith(prefix)

    def match_regex(entry):
        return bool(regex.match(entry.Key))

    def match_glob(entry):
        return fnmatch(entry.Key, glob)

    preds = []

    if prefix:
        prefix = prefix.lstrip("/")
        preds.append(match_prefix)

    if regex is not None:
        regex = re.compile(regex)
        preds.append(match_regex)

    if glob is not None:
        preds.append(match_glob)

    if len(preds) == 0:
        return lambda x: True
    elif len(preds) == 1:
        return preds[0]
    elif len(preds) == 2:
        p1, p2 = preds  # pylint:disable=unbalanced-tuple-unpacking
        return lambda e: p1(e) and p2(e)
    else:
        raise ValueError("regex and glob are mutually exclusive")


@click.command("s3-inventory-dump")
@click.option(
    "--inventory", "-i", type=str, help="URL pointing to manifest.json or one level up"
)
@click.option(
    "--prefix", type=str, help="Only print entries with Key starting with `prefix`"
)
@click.option("--regex", type=str, help="Only print entries matching regex")
@click.option("--aws-profile", type=str, help="Use non-default aws profile")
@click.option("--no-sign-request", is_flag=True, help="Do not sign AWS S3 requests")
@click.option(
    "--request-payer",
    is_flag=True,
    help="Needed when accessing requester pays public buckets",
)
@click.argument("glob", type=str, default="", nargs=1)
def cli(
    inventory,
    prefix,
    regex,
    glob,
    aws_profile,
    no_sign_request=None,
    request_payer=False,
) -> None:
    """List S3 inventory entries.

        prefix can be combined with regex or glob pattern, but supplying both
        regex and glob doesn't make sense.

    \b
    Example:
       s3-inventory s3://my-inventory-bucket/path-to-inventory/ '*yaml'

    """

    def entry_to_url(entry) -> str:
        return "s3://{e.Bucket}/{e.Key}".format(e=entry)

    opts = {}
    if request_payer:
        opts["RequestPayer"] = "requester"

    flush_freq = 100
    s3 = s3_client(profile=aws_profile, aws_unsigned=no_sign_request)

    if glob == "":
        glob = None

    if glob is not None and regex is not None:
        click.echo("Can not mix regex and shell patterns")
        sys.exit(1)

    if inventory is None:
        # TODO: read from config file
        inventory = "s3://dea-public-data-inventory/dea-public-data/dea-public-data-csv-inventory/"

    predicate = build_predicate(glob=glob, regex=regex, prefix=prefix)

    to_str = entry_to_url

    for i, entry in enumerate(list_inventory(inventory, s3=s3, **opts)):
        if predicate(entry):
            print(to_str(entry), flush=(i % flush_freq) == 0)


if __name__ == "__main__":
    cli()



import pytest
from click.testing import CliRunner
from pathlib import Path

from odc.apps.dc_tools.add_update_products import _get_product, _parse_csv
from odc.apps.dc_tools.add_update_products import cli as add_update_products_cli

TEST_DATA_FOLDER: Path = Path(__file__).parent.joinpath("data")
LOCAL_EXAMPLE: str = "example_product_list.csv"
PRODUCT_EXAMPLE: str = (
    "https://raw.githubusercontent.com/digitalearthafrica/"
    "config/master/products/esa_s2_l2a.odc-product.yaml"
)


def test_parse_local_csv(local_csv) -> None:
    local_contents = [x for x in _parse_csv(local_csv)]

    assert len(local_contents) == 13
    assert local_contents[0].name == "s2_l2a"


def test_parse_remote_csv(remote_csv) -> None:
    remote_contents = [x for x in _parse_csv(remote_csv)]
    assert len(remote_contents) == 13
    assert remote_contents[0].name == "s2_l2a"


def test_load_product_def(remote_product) -> None:
    products = _get_product(remote_product)

    assert products[0]["name"] == "s2_l2a"


def test_add_products(local_csv, odc_db, env_name) -> None:
    runner = CliRunner()
    # This will fail if requester pays is enabled
    result = runner.invoke(
        add_update_products_cli,
        [
            local_csv,
            "--update-if-exists",
            "--env",
            env_name,
        ],
    )
    print(f"CLI Output: {result.output}")
    assert result.exit_code == 0, f"Output: {result.output}"


@pytest.fixture
def remote_product():
    return PRODUCT_EXAMPLE


@pytest.fixture
def local_csv():
    return str(TEST_DATA_FOLDER / LOCAL_EXAMPLE)


@pytest.fixture
def remote_csv(httpserver, local_csv):
    httpserver.expect_request("/some.csv").respond_with_data(open(local_csv).read())
    yield httpserver.url_for("/some.csv")



import click
import tarfile
from odc.io.tar import add_txt_file, tar_mode
from odc.thredds import download_yamls, thredds_find_glob


@click.command("thredds-to-tar")
@click.option(
    "--thredds_catalogue",
    "-c",
    type=str,
    required=True,
    help="The THREDDS catalogue endpoint",
)
@click.option(
    "--skips",
    "-s",
    type=str,
    multiple=True,
    help="Pattern to ignore when THREDDS crawling",
)
@click.option(
    "--select",
    "-t",
    type=str,
    required=True,
    help="Target file pattern to match for yaml",
)
@click.option(
    "--workers",
    "-w",
    type=int,
    default=4,
    help="Number of thredds crawler workers to use",
)
@click.option(
    "--outfile", type=str, default="metadata.tar.gz", help="Sets the output file name"
)
def cli(thredds_catalogue, skips, select, workers, outfile) -> None:
    """Download Metadata from THREDDS server to tarball

    Example:

       \b
       Download files in directory that match `*yaml` and store them as a tar
        > thredds-to-tar -c "http://dapds00.nci.org.au/thredds/catalog/if87/2018-11-29/"
        -t ".*ARD-METADATA.yaml" -s '.*NBAR.*' -s '.*SUPPLEMENTARY.*'
         -s '.*NBART.*' -s '.*/QA/.*' -w 8 --outfile 2018-11-29.tar.gz

    """
    print(
        "Searching {thredds_catalogue} for matching files".format(
            thredds_catalogue=thredds_catalogue
        )
    )
    urls = thredds_find_glob(thredds_catalogue, skips, [select], workers)

    print("Found {0} metadata urls".format(str(len(urls))))

    yamls = download_yamls(urls, workers)

    # jam it all in a tar
    tar_opts = {
        "name": outfile,
        "mode": "w" + tar_mode(gzip=True, xz=True, is_pipe=False),
    }
    with tarfile.open(**tar_opts) as tar:
        for yaml in yamls:
            add_txt_file(tar=tar, content=yaml[0], fname=yaml[1])

    print("Done!")


if __name__ == "__main__":
    cli()



# Tests for the stac_api-to-dc CLI tool
from click.testing import CliRunner
from odc.apps.dc_tools.stac_api_to_dc import cli, item_to_meta_uri
from odc.apps.dc_tools.utils import MICROSOFT_PC_STAC_URI
from pystac import Item


def test_rewrite_urls(landsat_stac, odc_test_db_with_products):
    url_rewrite_tuple = (
        "https://dea-public-data-dev.s3-ap-southeast-2.amazonaws.com",
        "s3://dea-public-data-dev",
    )

    item = Item.from_dict(landsat_stac)

    print(item.self_href)

    _, uri, _ = item_to_meta_uri(
        item,
        odc_test_db_with_products,
        rename_product="ls8_c2l2_sr",
        url_string_replace=url_rewrite_tuple,
    )

    # https://dea-public-data-dev.s3-ap-southeast-2.amazonaws.com/
    # analysis-ready-data/ga_ls8c_ard_3/088/080/2020/05/25/ga_ls8c_ard_3-1-0_088080_2020-05-25_final.stac-item.json
    changed_uri = (
        "s3://dea-public-data-dev/analysis-ready-data/ga_ls8c_ard_3/088/080/2020/05/25/"
        "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.stac-item.json"
    )
    assert uri == changed_uri


def test_stac_to_dc_earthsearch(odc_test_db_with_products, env_name) -> None:
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--catalog-href=https://earth-search.aws.element84.com/v1/",
            "--bbox=5,15,10,20",
            "--limit=10",
            "--collections=sentinel-2-l2a",  # sentinel-s2-l2a-cogs is no longer available
            "--rename-product=s2_l2a",
            "--datetime=2020-08-01/2020-08-31",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert "Added 10 Datasets, failed 0 Datasets, skipped 0 Datasets" in result.output


def test_stac_to_dc_usgs(odc_test_db_with_products, env_name) -> None:
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--catalog-href=https://landsatlook.usgs.gov/stac-server/",
            "--bbox=5,15,10,20",
            "--limit=10",
            "--collections=landsat-c2l2-sr",
            "--datetime=2020-08-01/2020-08-31",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, f"Output: {result.output}"


def test_stac_to_dc_planetarycomputer(odc_test_db_with_products, env_name) -> None:
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            f"--catalog-href={MICROSOFT_PC_STAC_URI}",
            "--limit=1",
            "--collections=nasadem",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"Output: {result.output}"



"""Crawl Thredds for prefixes and fetch YAML's for indexing
and dump them into a Datacube instance
"""

import json
import logging
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Optional
import pystac

import click
from datacube import Datacube
from datacube.cfg import ODCEnvironment
from datacube.index.hl import Doc2Dataset
from datacube.metadata import stac2ds
from datacube.ui.click import environment_option, pass_config
from odc.apps.dc_tools.utils import (
    DatasetExists,
    allow_unsafe,
    archive_less_mature,
    fail_on_missing_lineage,
    index_update_dataset,
    publish_action,
    rename_product,
    skip_lineage,
    statsd_gauge_reporting,
    statsd_setting,
    transform_stac,
    update_flag,
    update_if_exists_flag,
    verify_lineage,
)
from odc.azure import download_blob, find_blobs
from sqlalchemy.exc import OperationalError, ProgrammingError


def stream_blob_urls(account_url, container_name, credential, blobs: List[str]):
    for blob in blobs:
        doc, uri, _ = download_blob(account_url, container_name, credential, blob)
        yield (json.loads(doc), uri)


def process_doc(
    doc,
    uri,
    dc,
    doc2ds,
    stac,
    update,
    update_if_exists,
    allow_unsafe,
    archive_less_mature,
    publish_action,
    rename_product,
) -> None:
    stac_doc = None
    if stac:
        stac_doc = doc
        if rename_product is not None:
            # This possibly should be possible for yaml loading too
            doc["properties"]["odc:product"] = rename_product
        doc = next(stac2ds([pystac.Item.from_dict(doc)]))
    index_update_dataset(
        doc,
        uri,
        dc,
        doc2ds,
        update=update,
        update_if_exists=update_if_exists,
        allow_unsafe=allow_unsafe,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        stac_doc=stac_doc,
    )


def dump_list_to_odc(
    account_url,
    container_name,
    credential,
    blob_urls: List[str],
    dc: Datacube,
    stac: Optional[bool] = False,
    update: Optional[bool] = False,
    update_if_exists: Optional[bool] = False,
    allow_unsafe: Optional[bool] = False,
    archive_less_mature: Optional[int] = None,
    publish_action: Optional[str] = None,
    rename_product: Optional[str] = None,
    **kwargs,
):
    doc2ds = Doc2Dataset(dc.index, **kwargs)

    # Do the indexing of all the things
    success = 0
    failure = 0
    skipped = 0

    sys.stdout.write("\rIndexing from Azure...\n")
    with ThreadPoolExecutor(max_workers=50) as executor:
        future_to_item = {
            executor.submit(
                process_doc,
                doc,
                uri,
                dc,
                doc2ds,
                stac,
                update,
                update_if_exists=update_if_exists,
                allow_unsafe=allow_unsafe,
                rename_product=rename_product,
                archive_less_mature=archive_less_mature,
                publish_action=publish_action,
            ): uri
            for doc, uri in stream_blob_urls(
                account_url, container_name, credential, blob_urls
            )
        }
        for future in as_completed(future_to_item):
            item = future_to_item[future]
            try:
                _ = future.result()
                success += 1
                if success % 10 == 0:
                    sys.stdout.write(f"\rAdded {success} datasets...")
            except DatasetExists:
                skipped += 1
            except Exception as e:  # pylint:disable=broad-except
                logging.exception("Failed to handle item %s with exception %s", item, e)
                failure += 1
    sys.stdout.write("\r")

    return success, skipped, failure


@click.command("azure-to-dc")
@environment_option
@pass_config
@update_flag
@update_if_exists_flag
@allow_unsafe
@transform_stac
@statsd_setting
@archive_less_mature
@publish_action
@click.option(
    "--account_url",
    "-a",
    "account_url",
    default=None,
    help=(
        "If you set the account URL, then you need a different"
        "kind of connection string to if you don't set one"
    ),
)
@click.argument("container_name", type=str, nargs=1)
@click.argument("credential", type=str, nargs=1)
@click.argument("prefix", type=str, nargs=1)
@click.argument("suffix", type=str, nargs=1)
@rename_product
@skip_lineage
@fail_on_missing_lineage
@verify_lineage
def cli(
    cfg_env: ODCEnvironment,
    update: bool,
    update_if_exists: bool,
    allow_unsafe: bool,
    stac: bool,
    statsd_setting: str,
    archive_less_mature: int,
    publish_action: str,
    account_url: str,
    container_name: str,
    credential: str,
    prefix: str,
    suffix: str,
    rename_product: str,
    skip_lineage: bool,
    fail_on_missing_lineage: bool,
    verify_lineage: bool,
) -> None:
    # Set up the datacube first, to ensure we have a connection
    try:
        dc = Datacube(env=cfg_env, app="azure-to-dc")
    except (OperationalError, ProgrammingError) as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)
    print(f"Opening AZ Container {container_name} on {account_url}")
    print(f"Searching on prefix '{prefix}' for files matching suffix '{suffix}'")
    yaml_urls = find_blobs(
        container_name, credential, prefix, suffix, account_url=account_url
    )

    # Consume generator and fetch YAML's
    added, skipped, failed = dump_list_to_odc(
        account_url,
        container_name,
        credential,
        yaml_urls,
        dc,
        stac=stac,
        update=update,
        update_if_exists=update_if_exists,
        allow_unsafe=allow_unsafe,
        archive_less_mature=archive_less_mature,
        publish_action=publish_action,
        rename_product=rename_product,
        skip_lineage=skip_lineage,
        fail_on_missing_lineage=fail_on_missing_lineage,
        verify_lineage=verify_lineage,
    )

    print(
        f"Added {added} Datasets, failed {failed} Datasets, skipped {skipped} Datasets"
    )
    if statsd_setting:
        statsd_gauge_reporting(
            added, ["app:azure_to_dc", "action:added"], statsd_setting
        )
        statsd_gauge_reporting(
            failed, ["app:azure_to_dc", "action:failed"], statsd_setting
        )


if __name__ == "__main__":
    cli()



import logging
from typing import Any, Dict, Optional, Tuple

import click
from datacube import Datacube
from datacube.index.hl import Doc2Dataset
from datacube.model import Dataset
from datacube.utils import changes, jsonify_document
from datadog import initialize, statsd
from odc.aws.queue import publish_to_topic
from datacube.metadata import stac2ds
from pystac import Item

from ._stac import ds_to_stac

MICROSOFT_PC_STAC_URI = "https://planetarycomputer.microsoft.com/api/stac/v1/"


class IndexingError(Exception):
    """
    Exception to raise for error during SQS to DC indexing/archiving
    """


class DatasetExists(Exception):
    """
    Exception to raise for error if dataset exists and not updating
    """


# A whole bunch of generic Click options
skip_lineage = click.option(
    "--skip-lineage",
    is_flag=True,
    default=False,
    help="Default is not to skip lineage. Set to skip lineage altogether.",
)

fail_on_missing_lineage = click.option(
    "--fail-on-missing-lineage",
    is_flag=True,
    help=(
        "Default is to permit unindexed/external lineage documents. "
        "Set flag to fail if lineage documents are not present in the database."
    ),
)

verify_lineage = click.option(
    "--verify-lineage",
    is_flag=True,
    default=False,
    help="Default is no verification. Set to verify parent dataset definitions.",
)

transform_stac = click.option(
    "--stac",
    is_flag=True,
    default=False,
    help="Expect STAC 1.0 metadata and attempt to transform to ODC EO3 metadata.",
)

transform_stac_absolute = click.option(
    "--absolute",
    is_flag=True,
    default=False,
    help="Use absolute paths from the STAC document.",
)

update_flag = click.option(
    "--update",
    is_flag=True,
    default=False,
    help="If set, update instead of add datasets.",
)

update_if_exists_flag = click.option(
    "--update-if-exists",
    is_flag=True,
    default=False,
    help="If the dataset or product already exists, update it instead of skipping it.",
)

allow_unsafe = click.option(
    "--allow-unsafe",
    is_flag=True,
    default=False,
    help="Allow unsafe changes to a dataset. Take care!",
)

skip_check = click.option(
    "--skip-check",
    is_flag=True,
    default=False,
    help="Assume file exists when listing exact file rather than wildcard.",
)

no_sign_request = click.option(
    "--no-sign-request",
    is_flag=True,
    default=False,
    help="Do not sign AWS S3 requests.",
)

request_payer = click.option(
    "--request-payer",
    is_flag=True,
    default=False,
    help="Needed when accessing requester pays public buckets.",
)

archive_less_mature = click.option(
    "--archive-less-mature",
    is_flag=False,
    flag_value=500,
    default=None,
    type=int,
    help=(
        "Archive existing any datasets that match product, "
        "time and region-code, but have lower dataset-maturity."
        "Note: An error will be raised and the dataset add will "
        "fail if a matching dataset with higher or equal dataset-maturity."
        "Can specify an of leniency for comparing timestamps, provided in milliseconds. "
        "Default value is 500ms."
    ),
)

publish_action = click.option(
    "--publish-action",
    type=str,
    default=None,
    nargs=1,
    help="SNS topic arn to publish indexing/archiving actions to.",
)

archive = click.option(
    "--archive",
    is_flag=True,
    default=False,
    help="Archive datasets instead of adding them.",
)

limit = click.option(
    "--limit",
    default=None,
    type=int,
    help="Stop indexing after n datasets have been indexed.",
)

bbox = click.option(
    "--bbox",
    type=str,
    default=None,
    help="Comma separated list of bounding box coords, lon-min, lat-min, lon-max, lat-max",
)

statsd_setting = click.option(
    "--statsd-setting",
    is_flag=False,
    default=None,
    help="statsd exporter hostname and port, i.e. prometheus-statsd-exporter:9125",
)

rename_product = click.option(
    "--rename-product",
    type=str,
    default=None,
    help=(
        "Name of product to overwrite collection(s) names, "
        "only one product name can overwrite, despite multiple collections "
    ),
)

url_string_replace = click.option(
    "--url-string-replace",
    type=str,
    default=None,
    help="Replace a string in the STAC API URLs, e.g., 'https://stac.example.com,s3://stac.example.org'",
)


def index_update_dataset(
    dataset: dict | Dataset,
    uri: str,
    dc: Datacube,
    doc2ds: Doc2Dataset | None,
    update: bool = False,
    update_if_exists: bool = False,
    allow_unsafe: bool = False,
    archive_less_mature: Optional[int] = None,
    auto_add_lineage: Optional[bool] = False,
    publish_action: Optional[str] = None,
    stac_doc: Optional[dict] = None,
):
    """
    Index and/or update a dataset.  Called by all the **_to_dc CLI tools.

    :param dataset: A dataset metadata dictionary, read from yaml or json, converted from STAC, etc.
    :param uri: The URI of the metadata and associated data.
    :param dc: A datacube object (carries a database index and potentially an active transaction).
    :param doc2ds: A Doc2Dataset object (metadata_type and product resolver)
    :param update: If true, allow update only.
    :param update_if_exists: If true allow insert or update.
    :param allow_unsafe: Allow unsafe (arbitrary) dataset updates.
    :param archive_less_mature: Enforce dataset maturity.
           * If None (the default), ignore dataset maturity.
           * If int, enforce dataset maturity by looking for existing datasets with same product, region_code and time
             values. If a less mature match is found, it is archived and replaced with the new dataset being inserted.
             If a match of the same or greater maturity is found a DatasetExists is raised.
             The integer value is used as the timedelta value for allowing a leniency when comparing
             timestamp values, for datasets where there is a slight discrepancy. Default is 500ms.
    :param publish_action: SNS topic arn to publish action to.
    :param stac_doc: STAC document for publication to SNS topic.
    :return: Returns nothing.  Raises an exception if anything goes wrong.
    """
    # Make sure we can create a dataset first
    if not isinstance(dataset, Dataset):
        try:
            if doc2ds is None:
                doc2ds = Doc2Dataset(dc.index)
            dataset, err = doc2ds(jsonify_document(dataset), uri)
        except ValueError as e:
            raise IndexingError(
                f"Exception thrown when trying to create dataset: '{e}'\n The URI was {uri}"
            ) from e
        if dataset is None:
            raise IndexingError(
                f"Failed to create dataset with error {err}\n The URI was {uri}"
            )

    with dc.index.transaction():
        # Process in a transaction
        archive_stacs = []
        added = False
        updated = False

        if isinstance(archive_less_mature, int) and publish_action:
            dupes = dc.index.datasets.find_less_mature(dataset, archive_less_mature)
            for dupe in dupes:
                archive_stacs.append(ds_to_stac(dupe))

        # Now do something with the dataset
        # Note that any of the exceptions raised below will rollback any archiving performed above.
        if dc.index.datasets.has(dataset.id):
            # Update
            if update or update_if_exists:
                # Set up update fields
                updates = {}
                if allow_unsafe:
                    updates = {tuple(): changes.allow_any}
                # Do the updating
                try:
                    dc.index.datasets.update(
                        dataset,
                        updates_allowed=updates,
                        archive_less_mature=archive_less_mature,
                    )
                    updated = True
                except ValueError as e:
                    raise IndexingError(
                        f"Updating the dataset raised an exception: {e}"
                    )
            else:
                raise DatasetExists(
                    f"Dataset {dataset.id} already exists, not indexing"
                )
        else:
            if update:
                # We're expecting to update a dataset, but it doesn't exist
                raise IndexingError("Can't update dataset because it doesn't exist.")
            # Everything is working as expected, add the dataset
            dc.index.datasets.add(
                dataset,
                with_lineage=auto_add_lineage,
                archive_less_mature=archive_less_mature,
            )
            added = True

    if publish_action:
        for arch_stac in archive_stacs:
            publish_to_topic(arn=publish_action, action="ARCHIVED", stac=arch_stac)

    if added:
        logging.info("New Dataset Added: %s", dataset.id)
        if publish_action:
            # if STAC was not provided, generate from dataset
            stac_doc = stac_doc if stac_doc else ds_to_stac(dataset)
            publish_to_topic(arn=publish_action, action="ADDED", stac=stac_doc)

    if updated:
        logging.info("Existing Dataset Updated: %s", dataset.id)


def statsd_gauge_reporting(value, tags=None, statsd_setting="localhost:8125") -> None:
    if tags is None:
        tags = []
    host = statsd_setting.split(":")[0]
    port = statsd_setting.split(":")[1]
    options = {"statsd_host": host, "statsd_port": port}
    initialize(**options)

    statsd.gauge("datacube_index", value, tags=tags)


def get_self_link(item: Item) -> str | None:
    uri = None
    for link in item.links:
        if link.rel == "self":
            uri = link.target

        # Override self with canonical
        if link.rel == "canonical":
            return link.target
    return uri


def item_to_meta_uri(
    item: Item,
    dc: Datacube,
    rename_product: Optional[str] = None,
    url_string_replace: tuple[str, str] | None = None,
) -> Tuple[Dataset, str | None, Dict[str, Any]]:
    if rename_product is not None:
        item.properties["odc:product"] = rename_product

    # If we need to modify URLs, do it for the main URL and the asset links
    uri = get_self_link(item)
    if url_string_replace is not None:
        old_url, new_url = url_string_replace

        if uri is not None:
            uri = uri.replace(old_url, new_url)

        for asset in item.assets.values():
            asset.href = asset.href.replace(old_url, new_url)

    # Try to the Datacube product for the dataset
    product_name = item.properties.get("odc:product", item.collection_id)
    product_name_sanitised = product_name.replace("-", "_")
    product = dc.index.products.get_by_name(product_name_sanitised)

    if product is None:
        logging.warning(
            "Couldn't find matching product for product name: %s",
            product_name_sanitised,
        )
        raise DatasetExists(
            f"Couldn't find matching product for product name: {product_name_sanitised}"
        )

    # Convert the STAC Item to a Dataset
    dataset = next(stac2ds([item], {"asset_absolute_paths": False}))
    # And assign the product ID
    dataset.product = product

    return (dataset, uri, item.to_dict(transform_hrefs=False))



from click.testing import CliRunner
from pathlib import Path

from odc.apps.dc_tools.fs_to_dc import cli as fs_to_dc_cli

TEST_DATA_FOLDER: Path = Path(__file__).parent.joinpath("data")


def test_fs_to_fc_yaml(test_data_dir, env_name, odc_test_db_with_products) -> None:
    runner = CliRunner()
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--stac",
            "--glob=**/NASADEM_HGT_s56w072.stac-item.json",
            "--env",
            env_name,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, f"Output: {result.output}"


def test_archive_less_mature(
    odc_db, env_name, test_data_dir, nrt_dsid, final_dsid
) -> None:
    dc = odc_db
    runner = CliRunner()

    # Index NRT dataset
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--glob=**/maturity-nrt.odc-metadata.yaml",
            "--archive-less-mature",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"fs-to-dc failed with output: {result.output}"
    assert dc.index.datasets.get(final_dsid) is None
    assert dc.index.datasets.get(nrt_dsid).archived_time is None

    # Index Final dataset (autoarchiving NRT)
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--glob=**/maturity-final.odc-metadata.yaml",
            "--archive-less-mature",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"fs-to-dc failed with output: {result.output}"
    assert dc.index.datasets.get(final_dsid).archived_time is None
    assert dc.index.datasets.get(nrt_dsid).archived_time is not None


def test_dont_archive_less_mature(
    odc_db, env_name, test_data_dir, nrt_dsid, final_dsid
) -> None:
    # no archiving should be done if --archive-less-mature is not set
    dc = odc_db
    runner = CliRunner()

    # Index NRT dataset
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--glob=**/maturity-nrt.odc-metadata.yaml",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"fs-to-dc failed with output: {result.output}"
    assert dc.index.datasets.get(final_dsid) is None
    assert dc.index.datasets.get(nrt_dsid).archived_time is None

    # Index Final dataset (autoarchiving NRT)
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--glob=**/maturity-final.odc-metadata.yaml",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"fs-to-dc failed with output: {result.output}"
    assert dc.index.datasets.get(final_dsid).archived_time is None
    assert dc.index.datasets.get(nrt_dsid).archived_time is None


def test_keep_more_mature(
    odc_db, env_name, test_data_dir, nrt_dsid, final_dsid
) -> None:
    dc = odc_db
    runner = CliRunner()

    # Index Final dataset
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--glob=**/maturity-final.odc-metadata.yaml",
            "--archive-less-mature",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"fs-to-dc failed with output: {result.output}"
    assert dc.index.datasets.get(nrt_dsid) is None
    assert dc.index.datasets.get(final_dsid).archived_time is None

    # Index NRT dataset (less mature - should be skipped)
    result = runner.invoke(
        fs_to_dc_cli,
        [
            test_data_dir,
            "--glob=**/maturity-nrt.odc-metadata.yaml",
            "--archive-less-mature",
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"fs-to-dc failed with output: {result.output}"
    assert dc.index.datasets.get(final_dsid).archived_time is None
    assert dc.index.datasets.get(nrt_dsid) is None



import pytest
from click.testing import CliRunner

from odc.apps.dc_tools.cop_dem_to_dc import get_dem_tile_uris, cli as cop_dem_to_dc_cli

PRODUCTS = ["cop_30", "cop_90"]


@pytest.fixture
def bbox() -> str:
    return "5,5,7,7"


@pytest.fixture
def bbox_africa() -> str:
    return "-26.359944882003788,-47.96476498374171,64.4936701740102,38.34459242512347"


@pytest.mark.parametrize("product", PRODUCTS)
def test_get_dem_tile_uris(bbox, product, odc_db) -> None:
    uris = list(get_dem_tile_uris(bbox, product))

    if product == "cop_30":
        assert uris[0][0] == (
            "https://copernicus-dem-30m.s3.eu-central-1.amazonaws.com/"
            "Copernicus_DSM_COG_10_N05_00_E005_00_DEM/Copernicus_DSM_COG_10_N05_00_E005_00_DEM.tif"
        )
    else:
        assert uris[0][0] == (
            "https://copernicus-dem-90m.s3.eu-central-1.amazonaws.com/"
            "Copernicus_DSM_COG_30_N05_00_E005_00_DEM/Copernicus_DSM_COG_30_N05_00_E005_00_DEM.tif"
        )

    assert len(uris) == 4


def test_complex_bbox(bbox_africa) -> None:
    uris = list(get_dem_tile_uris(bbox_africa, "cop_30"))

    assert len(uris) == 8004


# Test the actual process
@pytest.mark.parametrize("product", PRODUCTS)
def test_indexing_cli(bbox, product, odc_db, env_name) -> None:
    runner = CliRunner()
    result = runner.invoke(
        cop_dem_to_dc_cli,
        [
            "--add-product",
            "--bbox",
            bbox,
            "--product",
            product,
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert f"Product definition added for {product}" in result.output
    assert "Added 4 Datasets, failed 0 Datasets, skipped 0 Datasets" in result.output

    # Running a second time should skip the datasets
    result = runner.invoke(
        cop_dem_to_dc_cli,
        [
            "--add-product",
            "--bbox",
            bbox,
            "--product",
            product,
            "--env",
            env_name,
        ],
    )
    assert result.exit_code == 0, f"Output: {result.output}"
    assert f"Product definition added for {product}" in result.output
    assert "Added 0 Datasets, failed 0 Datasets, skipped 4 Datasets" in result.output
